/*
 *	Handle incoming frames
 *	Linux ethernet bridge
 *
 *	Authors:
 *	Lennert Buytenhek		<buytenh@gnu.org>
 *
 *	$Id: //WIFI_SOC/release/SDK_4_1_0_0/source/linux-2.6.21.x/net/bridge/br_input.c#1 $
 *
 *	This program is free software; you can redistribute it and/or
 *	modify it under the terms of the GNU General Public License
 *	as published by the Free Software Foundation; either version
 *	2 of the License, or (at your option) any later version.
 */

#include <linux/kernel.h>
#include <linux/netdevice.h>
#include <linux/inetdevice.h>
#include <linux/etherdevice.h>
#include <linux/netfilter_bridge.h>
#include <linux/string.h>

#include "br_private.h"

/* Bridge group multicast address 802.1d (pg 51). */
const u8 br_group_address[ETH_ALEN] = { 0x01, 0x80, 0xc2, 0x00, 0x00, 0x00 };

static void br_pass_frame_up(struct net_bridge *br, struct sk_buff *skb)
{
	struct net_device *indev;

	br->statistics.rx_packets++;
	br->statistics.rx_bytes += skb->len;

	indev = skb->dev;
	skb->dev = br->dev;

	NF_HOOK(PF_BRIDGE, NF_BR_LOCAL_IN, skb, indev, NULL,
		netif_receive_skb);
}


/* note: already called with rcu_read_lock (preempt_disabled) */
int br_handle_frame_finish(struct sk_buff *skb)
{
	const unsigned char *dest = eth_hdr(skb)->h_dest;
	struct net_bridge_port *p = rcu_dereference(skb->dev->br_port);
	struct net_bridge *br;
	struct net_bridge_fdb_entry *dst;
	int passedup = 0;

	if (!p || p->state == BR_STATE_DISABLED)
		goto drop;

	/* insert into forwarding database after filtering to avoid spoofing */
	br = p->br;
	br_fdb_update(br, p, eth_hdr(skb)->h_source);

	if (p->state == BR_STATE_LEARNING)
		goto drop;

	if (br->dev->flags & IFF_PROMISC) {
		struct sk_buff *skb2;

		skb2 = skb_clone(skb, GFP_ATOMIC);
		if (skb2 != NULL) {
			passedup = 1;
			br_pass_frame_up(br, skb2);
		}
	}


	if (is_multicast_ether_addr(dest)) {
#ifdef CONFIG_BRIDGE_IGMP_REPORT_NO_FLOODING
		#include <linux/ip.h>
		#include <linux/igmp.h>
		struct ethhdr *eth = eth_hdr(skb);
		struct iphdr *ih;
		struct igmphdr *igmph;

		if(dest[0] != 0x01 || dest[1] != 0x00 || dest[2] != 0x5e || (dest[3] > 0x7f))
			goto out_igmp;

		if(eth->h_proto == htons(ETH_P_IP)){
			if(skb->len < (sizeof(struct iphdr) + sizeof(struct igmphdr)))
				goto out_igmp;

			ih = (struct iphdr *)skb->h.raw;
			if(ih->protocol != IPPROTO_IGMP)
				goto out_igmp;

			igmph = (struct igmphdr *)((unsigned char *)skb->h.raw + (ih->ihl<<2)  );
			if(	igmph->type == IGMP_HOST_MEMBERSHIP_REPORT ||
				igmph->type == IGMPV2_HOST_MEMBERSHIP_REPORT ||
				igmph->type == IGMPV3_HOST_MEMBERSHIP_REPORT){
				if(!passedup)
					br_pass_frame_up(br, skb);
				goto out;
			}
		}
out_igmp:
#endif
		br->statistics.multicast++;
		br_flood_forward(br, skb, !passedup);
		if (!passedup)
			br_pass_frame_up(br, skb);
		goto out;
	}

	dst = __br_fdb_get(br, dest);
	if (dst != NULL && dst->is_local) {
		if (!passedup)
			br_pass_frame_up(br, skb);
		else
			kfree_skb(skb);
		goto out;
	}

	if (dst != NULL) {
		br_forward(dst->dst, skb);
		goto out;
	}

	br_flood_forward(br, skb, 0);

out:
	return 0;
drop:
	kfree_skb(skb);
	goto out;
}

/* note: already called with rcu_read_lock (preempt_disabled) */
static int br_handle_local_finish(struct sk_buff *skb)
{
	struct net_bridge_port *p = rcu_dereference(skb->dev->br_port);

	if (p && p->state != BR_STATE_DISABLED)
		br_fdb_update(p->br, p, eth_hdr(skb)->h_source);

	return 0;	 /* process further */
}

/* Does address match the link local multicast address.
 * 01:80:c2:00:00:0X
 */
static inline int is_link_local(const unsigned char *dest)
{
	return memcmp(dest, br_group_address, 5) == 0 && (dest[5] & 0xf0) == 0;
}
#define AIRMOBI_ACCESS_BY_HOST_NAME
#ifdef AIRMOBI_ACCESS_BY_HOST_NAME
typedef struct _AMDnsHeader {
        unsigned short id;
        unsigned short flag;
        unsigned short ques;
        unsigned short answ;
        unsigned short auth;
        unsigned short addrrs;
} AMDnsHeader;

typedef struct _AMDnsFixedHeader{
    unsigned short type;
    unsigned short classtype;
    unsigned int   ttl;
    unsigned short rdlengh;
} AMDnsFixedHeader;

/*
    pName: ʽ rfc1035: 3.1. Name space definitions
    pEnd:    pNmaeıַ߽
    return:  pNameĳ
 */
int AM_dns_DomainNameLen(char *pName, char *pEnd)
{
    char        *pRead = pName;
    int          iNameLen = 0;
    unsigned int uiLabelLen;
    int          bFinish = 0;
    
    for(;pRead < pEnd;)
    {
        if((*pRead) & 0xC0)
        {
            iNameLen += 2; //ѹʽLabel/
            pRead    += 2;
            bFinish   = 1;
            break;
        }
        else
        {
            uiLabelLen = (*(unsigned char *)pRead) & 0x3F;
            iNameLen += (1 + uiLabelLen);
            pRead     += (1 + uiLabelLen);
            if(!uiLabelLen)
            {
                bFinish = 1;
                break;  
            }
        }
    }
    if(pRead < pEnd)
    {
        return iNameLen;
    }
    else if(pRead == pEnd)
    {
        if(bFinish)
        {
            return iNameLen; //泤ȸոպ
        }
        else
        {
            return -1; //泤Ȳ
        }
    }
    else
    {
        return -1;
    }

}

/*
    һDNSϢAnswerĳһRRеNameǷƥض
    pDnsMsg:   DNSԭʼϢ
    pRRName: pDnsMsgϢAnswerĳһRRNameֶεַ
    pEnd:        pDnsMsgϢı߽磬()ñ߽ĵַɷʣ
                   pDnsMsgpRRNameԽ籣
    sHostName: 0βַ"www.cctairmobi.com"

    return:      true:ȷƥ䣬false:ƥ

  */
int AM_dns_TryMatchHostName(char *pDnsMsg, char *pRRName, char *pEnd, char *sHostName)
{
    unsigned short usOffset;
    char          *pReadLabel    = pRRName;
    char          *pReadHostName = sHostName;
    unsigned int   uiHostSegLen;
    int            bMatchSeg = 0; 
    unsigned int   uiLabelLen;

    for(;;)
    {
        //׼HostNameһ
        
        uiHostSegLen = 0;
        if(!pReadHostName[uiHostSegLen]) //sHostNameжƥ
        {
            break;
        }
        while(pReadHostName[uiHostSegLen] && pReadHostName[uiHostSegLen] != '.')
        {
            uiHostSegLen++;
        }
        
        //׼DNS Ϣеƥlabel
        
        if(pReadLabel >= pEnd)
        {
            bMatchSeg = 0;
            break;
        }
        if(((*pReadLabel) & 0xC0) == 0xC0)
        {
            //ѹʽ
            if(pReadLabel + 2 <= pEnd)
            {
                usOffset = ntohs(*((unsigned short *)pReadLabel)) & 0x3FFF;
                pReadLabel = pDnsMsg + usOffset;
            }
            else
            {
                bMatchSeg = 0; //Ϣʽȷ
                break;
            }
        }
        //else ѹʽlabel

        //Ƚϸ
        
        if(pReadLabel >= pEnd)
        {
            bMatchSeg = 0;
            break;
        }
        uiLabelLen = (*(unsigned char *)pReadLabel) & 0x3F;
        pReadLabel++;
        if(pReadLabel + uiLabelLen > pEnd)
        {
            bMatchSeg = 0;
            break;
        }
        if(uiLabelLen != uiHostSegLen)
        {
            bMatchSeg = 0;
            break;
        }
        if(!strnicmp(pReadLabel, pReadHostName, uiLabelLen))
        {
            bMatchSeg = 1; // ƥ
            pReadLabel += uiLabelLen;
        }
        else
        {
            bMatchSeg = 0;
            break;
        }

        //pReadHostName
        
        pReadHostName += uiHostSegLen;
        if(*pReadHostName == '.')
        {
            pReadHostName++;
        }
    }
    
    if(bMatchSeg && (pReadLabel < pEnd) && !(*pReadLabel))
    {
        return 1;
    }
    else
    {
        return 0;
    }
}

/*
    pMsgBuf:     ԭʼDNSϢ
    pRRName:   pMsgBufеAnswers һRR (rfc1035 3.2. RR definitions)
    pEnd:          pMsgBufpRRNameıַ߽
    sHostName: 0βַ"www.cctairmobi.com"
    bIsReplaced: Σtrue: ɹ滻,0: δ滻
    
    return: RRĳ,<0:
*/
int AM_dns_CheckAndReplaceHostAddr(char *pMsgBuf, char *pRRName, char *pEnd, char *sHostName, int *pbIsReplaced)
{
    int   iDomainNameLen;
    int   iReadLen;
    int   iAddrLen;
    int   iRrLen = 0;
    char *ptr = pRRName;
    AMDnsFixedHeader *pRrFixedHdr;

    *pbIsReplaced = 0;

    iDomainNameLen = AM_dns_DomainNameLen(ptr, pEnd);
    if(iDomainNameLen > 0)
    {
        ptr    += iDomainNameLen;
        iRrLen += iDomainNameLen;
    }
    else
    {
        return -1;
    }

    iReadLen = 10;   // ʹ sizeof(AMDnsFixedHeader), Աֽڶ
    if(ptr + iReadLen > pEnd)
    {
        return -1;
    }
    
    pRrFixedHdr = (AMDnsFixedHeader *)ptr;
    ptr        += iReadLen;
    iRrLen     += iReadLen;

    iAddrLen = ntohs(pRrFixedHdr->rdlengh);

    if( (ntohs(pRrFixedHdr->type)      == 1)  // rfc1035 1: Host Address
      && (ntohs(pRrFixedHdr->classtype) == 1)  // rfc1035 1: the Internet
      && (iAddrLen                      == 4) // leng of ipv4 address
      && AM_dns_TryMatchHostName(pMsgBuf, pRRName, pEnd, sHostName))
    {
        if(ptr + iAddrLen <= pEnd)
        {
			struct net_device *pNetDev = dev_get_by_name("br0");
			if(pNetDev && pNetDev->ip_ptr && ((struct in_device *)pNetDev->ip_ptr)->ifa_list)
			{
	            pRrFixedHdr->ttl = 0;               // not allowed for cache
	            memcpy(ptr, &(((struct in_device *)pNetDev->ip_ptr)->ifa_list->ifa_address), iAddrLen);  //Replace the Address
	           *pbIsReplaced = 1;
			    if( pNetDev!=NULL )
			    {
					dev_put(pNetDev);
				}
	            return iRrLen + iAddrLen;
			}
			else
			{
				if( pNetDev!=NULL )
			    {
					dev_put(pNetDev);
				}
				return -1;
			}
        }
        else
        {
            return -1;
        }
    }
    else
    {
        return iRrLen + ntohs(pRrFixedHdr->rdlengh);
    }

}


/*
      Replace inet host addresses in dns response message
      
      pMsgBuf:      raw dns message
      iMsgLen:      length of dns message
      sHostName: 0βַ"www.cctairmobi.com"
      return:        true: updated, false: no updated
  */
int AM_dns_Update(char *pMsgBuf, int iMsgLen, char *sHostName)
{
	char *pRead;
    char *pEnd; 
    int  iReadLen;
    int  iDomainNameLen;
    int  iIdx;
    int  iRet;
	int  iAnsCount = 0;
    int  iQuestionCount = 0;

    int  bHaveReplaced = 0;
    int  bTmp;
    AMDnsHeader *pMsgHdr;

    pRead = pMsgBuf;
    pEnd  = pMsgBuf + iMsgLen; //addresses smaller than this pointer are avaliable to access

    //DNS Message Header
    
    iReadLen = 12; //ʹsizeof(AMDnsHeader)Աṹ
    if(pRead + iReadLen > pEnd)
    {
        return 0;  // buf len short than DNS header
    }
   
	pMsgHdr             = (AMDnsHeader *)pRead;
    iQuestionCount      = ntohs(pMsgHdr->ques);
	iAnsCount           = htons(pMsgHdr->answ);
    
	pRead += iReadLen;  //skip dns message header

    //Skip all Questions
    
    for(iIdx = 0; iIdx < iQuestionCount && pRead < pEnd; iIdx++)
    {
        //QNAME
        iDomainNameLen = AM_dns_DomainNameLen(pRead, pEnd);
        if(iDomainNameLen > 0)
        {
            pRead += iDomainNameLen;
        }
        else
        {
            return 0;  // Query
        }

        //QTYPE and QCLASS
        iReadLen = 4;
        if(pRead + iReadLen > pEnd)
        {
            return 0;  // Query
        }
        pRead += iReadLen;
    }

    if(iIdx != iQuestionCount)
    {
        return 0; //Ϣʽ
    }

    //ƥ䲢滻IP
    for(iIdx = 0; iIdx < iAnsCount; iIdx++)
    {
        iRet = AM_dns_CheckAndReplaceHostAddr(pMsgBuf, pRead, pEnd, sHostName, &bTmp); 
        if(iRet >= 0)
        {
            if(bTmp)
            {
                bHaveReplaced = 1;
            }
            pRead += iRet;
        }
        else
        {
            return 0;
        }
    }
	return bHaveReplaced;
}


#endif //AIRMOBI_ACCESS_BY_HOST_NAME

/*
 * Called via br_handle_frame_hook.
 * Return 0 if *pskb should be processed furthur
 *	  1 if *pskb is handled
 * note: already called with rcu_read_lock (preempt_disabled)
 */
int br_handle_frame(struct net_bridge_port *p, struct sk_buff **pskb)
{
	
#ifdef AIRMOBI_ACCESS_BY_HOST_NAME
	
		struct ethhdr  *pEthHdr;
		unsigned char  *pIpHdr;
		unsigned char  *pUdpHdr;
		unsigned short	usNetProt;
		unsigned char	cIPVer;
		unsigned char	cTransProt;
		unsigned short	usSrcPort;
	 
#define AM_ETH_HDR_LEN 14
#define AM_IP4_HDR_LEN(ipHdr) ((*((char *)ipHdr) &0x0F)*4)
#define AM_UDP_HDR_LEN 8
#endif //AIRMOBI_ACCESS_BY_HOST_NAME

	struct sk_buff *skb = *pskb;
	const unsigned char *dest = eth_hdr(skb)->h_dest;

	if (!is_valid_ether_addr(eth_hdr(skb)->h_source))
		goto err;

#ifdef AIRMOBI_ACCESS_BY_HOST_NAME

    pEthHdr = eth_hdr(skb);
	pIpHdr  = (skb->mac.raw + AM_ETH_HDR_LEN);
	pUdpHdr = (skb->mac.raw + AM_ETH_HDR_LEN + AM_IP4_HDR_LEN(pIpHdr));

		if(pUdpHdr + AM_UDP_HDR_LEN <= skb->tail)
    {
        usNetProt = ntohs(pEthHdr->h_proto);
        cIPVer = (*pIpHdr >> 4) & 0x0F;                   // ip->versoin
        cTransProt = *(pIpHdr + 9);                        // ip->protocol
        usSrcPort = ntohs(*(unsigned short *)(pUdpHdr));     //udp->src port

        if( usNetProt == ETH_P_IP //ip
           && cIPVer == 4         //ip v4
           && cTransProt == 17    //udp
           && usSrcPort == 53)       //dns
        {
        
#if 0

            struct net_device *pNetDev = dev_get_by_name("br0");
			struct in_device *pInDev;
			if(pNetDev)
			{
	            printk(KERN_INFO"bridge device:%X, name[%s]  \n",pNetDev, pNetDev->name);
	            pInDev = pNetDev->ip_ptr;
	            printk(KERN_INFO"bridge device->ip_ptr->ifa_list:%X  \n",pInDev->ifa_list);
	            if(pInDev->ifa_list)
	            {
	                printk(KERN_INFO"bridge device->ip_ptr->ifa_list->address:%X  \n",((struct in_device *)pNetDev->ip_ptr)->ifa_list->ifa_address);
	            }
			}
#endif

            if(AM_dns_Update((pUdpHdr + AM_UDP_HDR_LEN)                  
                              ,skb->tail - (pUdpHdr + AM_UDP_HDR_LEN)
                              ,"c.cctairmobi.com"))
            {
                pUdpHdr[6] = 0; //udp checksum
                pUdpHdr[7] = 0; //udp checksum     
                printk(KERN_INFO"Access by Host Name\n");
            }
			
        }

    }
#endif //AIRMOBI_ACCESS_BY_HOST_NAME

	if (unlikely(is_link_local(dest))) {
		skb->pkt_type = PACKET_HOST;
		return NF_HOOK(PF_BRIDGE, NF_BR_LOCAL_IN, skb, skb->dev,
			       NULL, br_handle_local_finish) != 0;
	}

	if (p->state == BR_STATE_FORWARDING || p->state == BR_STATE_LEARNING) {
		if (br_should_route_hook) {
			if (br_should_route_hook(pskb))
				return 0;
			skb = *pskb;
			dest = eth_hdr(skb)->h_dest;
		}

		if (!compare_ether_addr(p->br->dev->dev_addr, dest))
			skb->pkt_type = PACKET_HOST;

		NF_HOOK(PF_BRIDGE, NF_BR_PRE_ROUTING, skb, skb->dev, NULL,
			br_handle_frame_finish);
		return 1;
	}

err:
	kfree_skb(skb);
	return 1;

#ifdef AIRMOBI_ACCESS_BY_HOST_NAME

    #undef AM_ETH_HDR_LEN
    #undef AM_IP4_HDR_LEN
    #undef AM_UDP_HDR_LEN
#endif //AIRMOBI_ACCESS_BY_HOST_NAME
	
}
