#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <fcntl.h>
#include <signal.h>
#include <sys/ioctl.h>
#include <linux/autoconf.h>
#include "ralink_gpio.h"

#define LED_ON				1
#define LED_OFF				0

#define GPIO_DEV	"/dev/gpio"

typedef enum {
	LED_UNKNOWN_ACT,
	LED_SET_ON,
	LED_SET_OFF
}TemLedAction;

static struct {
	const char *pActionStr;
	TemLedAction emAction;
}g_led_action[] = {
	{"on", LED_SET_ON},
	{"off", LED_SET_OFF}
};

#define	ARRAY_SIZEOF(array)	sizeof(array)/sizeof(array[0])
#ifdef	SINGLE_COMMAND
#define	LED_PARMS_IDX	(1)	/* if command format is: "led on/off", thus params behind offset 0. */
#else
#define	LED_PARMS_IDX	(2)	/* if command format is: "<prefix_command> led on/off", thus params behind offset 1. */
#endif

#define	LED_PWR_GPIO	(8)
#define	LED_SECU_GPIO	(13)
#define	LED_WPS_GPIO	(14)
static int led_gpio_define[] = { LED_PWR_GPIO, LED_SECU_GPIO, LED_WPS_GPIO };

/*
 * configure LED blinking with proper frequency (privatly use only)
 *   on: number of ticks that LED is on
 *   off: number of ticks that LED is off
 *   blinks: number of on/off cycles that LED blinks
 *   rests: number of on/off cycles that LED resting
 *   times: stop blinking after <times> times of blinking
 * where 1 tick == 100 ms
 */
static int gpioLedSet(int gpio, unsigned int on, unsigned int off,
		unsigned int blinks, unsigned int rests, unsigned int times)
{
	int fd;
	ralink_gpio_led_info led;

	//parameters range check
	if (gpio < 0 || gpio >= RALINK_GPIO_NUMBER ||
			on > RALINK_GPIO_LED_INFINITY ||
			off > RALINK_GPIO_LED_INFINITY ||
			blinks > RALINK_GPIO_LED_INFINITY ||
			rests > RALINK_GPIO_LED_INFINITY ||
			times > RALINK_GPIO_LED_INFINITY) {
		return -1;
	}
	led.gpio = gpio;
	led.on = on;
	led.off = off;
	led.blinks = blinks;
	led.rests = rests;
	led.times = times;

	fd = open(GPIO_DEV, O_RDONLY);
	if (fd < 0) {
		perror(GPIO_DEV);
		return -1;
	}
	if (ioctl(fd, RALINK_GPIO_LED_SET, &led) < 0) {
		perror("ioctl");
		close(fd);
		return -1;
	}
	close(fd);
	return 0;
}

static int ledAlways(int gpio, int on)
{
	if (on)
		return gpioLedSet(gpio, RALINK_GPIO_LED_INFINITY, 0, 1, 1, RALINK_GPIO_LED_INFINITY);
	else
		return gpioLedSet(gpio, 0, RALINK_GPIO_LED_INFINITY, 1, 1, RALINK_GPIO_LED_INFINITY);
}

/* preset power led status. */
static void led_power_prepare(TemLedAction emAct)
{
	int fd;

	if ( (emAct != LED_SET_ON) || (emAct != LED_SET_OFF) )
		return;

	fd = open(GPIO_DEV, O_RDONLY);
	if (fd < 0) {
		perror(GPIO_DEV);
		return;
	}

	/* power led prepare */
	if ( ioctl(fd, RALINK_GPIO_SET_DIR_IN, RALINK_GPIO(9)) < 0 )
		perror("pwr_led ioctl");
	close(fd);

	return;
}

#define	RESULT_OK	"ok"
#define	RESULT_FAILED	"fail"
static void led_action(TemLedAction emAct)
{
	int i;

	led_power_prepare(emAct);
	switch(emAct) {
		case LED_SET_ON:
			/* switch on wlan led */
			if (-1 == system("iwpriv ra0 set RadioOn=1") )
				goto LED_ACTION_FAILED;
			/* other leds on */
			for (i=0; i<ARRAY_SIZEOF(led_gpio_define); i++)
				if (0 != ledAlways(led_gpio_define[i], LED_ON) )
					goto LED_ACTION_FAILED;
			break;
		case LED_SET_OFF:
			/* switch off wlan led */
			if (-1 == system("iwpriv ra0 set RadioOn=0") )
				goto LED_ACTION_FAILED;
			/* other leds off */
			for (i=0; i<ARRAY_SIZEOF(led_gpio_define); i++)
				if (0 != ledAlways(led_gpio_define[i], LED_OFF) )
					goto LED_ACTION_FAILED;
			break;
		default:;
	} /* end switch(emAct) */

	/* all done ok */
	printf(RESULT_OK);
	return;

LED_ACTION_FAILED:
	printf(RESULT_FAILED);
	return;
}

/* Usage: led <on/off> */
#ifndef	SINGLE_COMMAND
int led_main(int argc, char *argv[])
#else
int main(int argc, char *argv[])
#endif
{
	TemLedAction action = LED_UNKNOWN_ACT;
	int i;

	if (argc < 2)
		return -1;

	for (i = 0; i < ARRAY_SIZEOF(g_led_action); i++){
		if (0 == strncmp(argv[LED_PARMS_IDX], g_led_action[i].pActionStr,
			strlen(g_led_action[i].pActionStr)) ){
			action = g_led_action[i].emAction;
			break;
		}
	}

	//unknown led action. show result: failed
	if (LED_UNKNOWN_ACT == action){
		printf(RESULT_FAILED);
		return 0;
	}

	led_action(action);

	return 0;
}
