/*
 * link chain utility functions
 *
 * Copyright (C) 2011, WDK Corporation
 * All Rights Reserved.
 *
 * THIS SOFTWARE IS OFFERED "AS IS", AND WDK GRANTS NO WARRANTIES OF ANY
 * KIND, EXPRESS OR IMPLIED, BY STATUTE, COMMUNICATION OR OTHERWISE. WDK
 * SPECIFICALLY DISCLAIMS ANY IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS
 * FOR A SPECIFIC PURPOSE OR NONINFRINGEMENT CONCERNING THIS SOFTWARE.
 *
 * $Id: lutils.c, by zhaoyanming 2011/06/02 Exp $
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <lutils.h>

#ifdef DEBUG
#define dpf(fmt, args...) printf("(%s:%d)=> " fmt, __FUNCTION__, __LINE__, ## args)
#else
#define dpf(fmt, args...)
#endif 

class_link_t *ltls_create_link_object( add_entry_f *addf,
                                       match_entry_f *machf,
                                       del_entry_f *delf )
{
    class_link_t *p_obj = NULL;

    if( addf == NULL || machf == NULL || delf == NULL )
    {
        dpf( "argument error.\n" );
        return NULL;
    }

    p_obj = malloc( sizeof( class_link_t ) );
    if( p_obj == NULL )
    {
        printf( "create_link_object: malloc memery fail.\n" );
        return NULL;
    }

    memset( p_obj, 0, sizeof( class_link_t ) );
    p_obj->i_fresh_flag = 1;
    p_obj->add_func = addf;
    p_obj->match_func = machf;
    p_obj->del_func = delf;

    return p_obj;
}


int ltls_add_node( class_link_t *p_obj, void *p_data )
{
    node_t *pn = NULL, *p_new = NULL;
    int i_result = 0;

    if( p_obj == NULL )
    {
        dpf( "argument error.\n" );
        return -1;
    }

    pn = p_obj->p_head;
    if( p_obj->add_func == NULL )
    {
        printf( "add fucntion isn't empty.\n" );
        goto end;
    }

    p_new = ( node_t * )malloc( sizeof( node_t ) );
    if( p_new == NULL )
    {
        printf( "malloc memery fail.\n" );
        i_result = -1;
        goto end;
    }

    if( pn == NULL )
    {
        i_result = p_obj->add_func( p_new, p_data ) ;
        if( i_result != 0 )
        {
            free( p_new );
            i_result = -1;
            goto end;
        }
        p_new->p_next = NULL;
        p_obj->p_head = p_new;
        p_obj->p_end = p_new;
        p_obj->p_curr = p_new;
        p_obj->i_node_sum++;
    }
    else
    {
        while( pn )
        {
            if( pn->p_next == NULL )
            {
                i_result = p_obj->add_func( p_new, p_data ) ;
                if( i_result != 0 )
                {
                    free( p_new );
                    i_result = -1;
                    goto end;
                }
                p_new->p_next = NULL;
                pn->p_next = p_new;
                p_obj->p_end = p_new;
                p_obj->i_node_sum++;

                break;
            }
            pn = pn->p_next;
        }
    }
end:
    return i_result;
}

node_t *ltls_get_node( class_link_t *p_obj, void *p_entry, void *p_key )
{
    node_t *pn = NULL, *p_tgt = NULL;

    if( p_obj == NULL || p_obj->p_head == NULL
        || p_obj->match_func == NULL
        || ( p_entry == NULL && p_key == NULL ) )
    {
        dpf( "argument error.\n" );
        return NULL;
    }

    pn = p_obj->p_head;
    while( pn )
    {
        if( p_obj->match_func( pn, p_entry, p_key ) )
        {
            p_tgt = pn;
            break;
        }

        pn = pn->p_next;
    }

    return p_tgt;
}

void *ltls_get_entry( class_link_t *p_obj, void *p_entry, void *p_key )
{
    node_t *p_node = NULL;

    if( p_obj == NULL )
    {
        dpf( "argument error.\n" );
        return NULL;
    }
    p_node = ltls_get_node( p_obj, p_entry, p_key );
    return (p_node==NULL)?NULL:(p_node->p_entry);
}


int ltls_fresh_object( class_link_t *p_obj )
{
    if( p_obj == NULL )
    {
        dpf( "argument error.\n" );
        return -1;
    }

    p_obj->i_node_sum = 0;
    p_obj->p_curr = p_obj->p_head;
    while( p_obj->p_curr )
    {
        p_obj->i_node_sum++;
        p_obj->p_curr = p_obj->p_curr->p_next;
    }

    p_obj->i_fresh_flag = 1;
    p_obj->p_curr = p_obj->p_head;

    return 0;
}

node_t *ltls_get_next_node( class_link_t *p_obj )
{
    if( p_obj == NULL )
    {
        dpf( "argument error.\n" );
        return NULL;
    }

    if( p_obj->i_fresh_flag != 0 )
    {
        p_obj->i_fresh_flag = 0;
        return p_obj->p_curr;
    }

    p_obj->p_curr = p_obj->p_curr->p_next;
    return p_obj->p_curr;
}

void *ltls_get_next_entry( class_link_t *p_obj )
{
    node_t *p_node = NULL;

    if( p_obj == NULL )
    {
        dpf( "argument error.\n" );
        return NULL;
    }

    p_node = ltls_get_next_node( p_obj );
    return (p_node==NULL)?NULL:(p_node->p_entry);
}

int ltls_del_node( class_link_t *p_obj, node_t *p_node )
{
    node_t *pn = NULL, *p_priv = NULL;

    if( p_obj == NULL || p_obj->p_head == NULL
        || p_node == NULL || p_obj->del_func == NULL )
    {
        dpf( "argument error.\n" );
        return -1;
    }

    pn = p_obj->p_head;
    p_priv = p_obj->p_head;
    while( pn )
    {
        if( pn == p_node )
        {
            if( pn == p_obj->p_curr )
            {
                if( pn->p_next != NULL )
                    p_obj->p_curr = pn->p_next;
                else
                    p_obj->p_curr = p_priv;
            }
            if( pn == p_obj->p_head )
                p_obj->p_head = pn->p_next;
            else
                p_priv->p_next = pn->p_next;
            if( pn == p_obj->p_end )
                p_obj->p_end = p_priv;
            p_obj->del_func( pn->p_entry );
            p_obj->i_node_sum--;
            free( pn );

            break;
        }

        p_priv = pn;
        pn = pn->p_next;
    }

    return 0;
}

int ltls_free_nodes( class_link_t *p_obj )
{
    node_t *pn = NULL, *p_node = NULL;

    if( p_obj == NULL || p_obj->del_func == NULL || p_obj->p_head == NULL )
    {
        dpf( "argument error.\n" );
        return -1;
    }

    pn = p_obj->p_head;
    while( pn )
    {
        p_node = pn->p_next;
        p_obj->del_func( pn->p_entry );
        free( pn );

        pn = p_node;
    }

    return 0;
}

class_link_t *ltls_destory_object( class_link_t *p_obj )
{
    if( p_obj == NULL )
    {
        dpf( "argument error.\n" );
        return NULL;
    }

    ltls_free_nodes( p_obj );
    free( p_obj );

    return NULL;
}

